#!/usr/bin/env python3


import os
import re
import sys
from argparse import ArgumentParser, Namespace
from asyncio import StreamReader, create_subprocess_exec, gather, run, CancelledError
from asyncio.subprocess import PIPE
from collections.abc import Iterator,   Sequence
from contextlib import contextmanager
from pathlib import Path
from datetime import datetime
from typing import TextIO
import yaml
from contextlib import suppress

import aiofiles



def load_filter_pattern(file_path: str|Path):
    with suppress(FileNotFoundError):
        with Path(file_path).expanduser().open() as config_file:
            file_content = yaml.load(config_file, Loader=yaml.BaseLoader)
            return "|".join(rf".*{pattern}.*" for pattern in (
                pattern.replace("~", str(Path("~").expanduser())) for pattern in file_content["exclude"]))


def extract_paths(stream):
    async for line in stream:
        if (match := re.match(rf'^\d* openat\(([A-Z_]+|\d+), (".*"), ([A-Z_\|]+).*\s=\s((\?|-?\d+)( [A-Z_]+)?( \(.*\))?)\n$', line)):
            location, filename, flags, _result, result_nr, *_rest = match.groups()
        elif re.match(r"^\d* ((<\.\.\. .* resumed>)|(\?\?\?\?\( <unfinished \.\.\.>)|(\+\+\+ )|(\-\-\- )).*$", line):
            pass
        else:
            print(f"strace: {line}")


async def process_strace_lines(filename: Path, out_file) -> None:
    """For testability: provides content of a file to process_strace()"""
    filter_pattern = load_filter_pattern("~/.config/procmon-exclude.yaml")
    accessed_files = set()
    last_line = ""
    try:
        async with aiofiles.open(filename) as afp:
            async for path_entry in extract_paths(afp):
                out_file.write(f"{filename} | {line}\n")
                if not re.match(filter_pattern, filename):
                    print(f"openat: {location} {flags}: {filename}")
                    if filename not in accessed_files:
                        accessed_files.add(filename)
                    else:
                        # print(f"openat: {flags}: {filename}")
                        pass
    finally:
        pass
        #for file in sorted(accessed_files):
        #    print(file.replace(str(Path("~").expanduser()),"~"))


async def buffer_stream(stream: StreamReader, out_file: TextIO) -> None:
    """Records a given stream to a buffer line by line along with the source"""
    while line := (await stream.readline()).decode():
        out_file.write(line)


@contextmanager
def strace_output_path(path: Path | None = None) -> Iterator[Path]:
    """Wraps optional creation of named pipe and sanatizes @path argument"""
    result = path or Path("myfifo")
    result.unlink(missing_ok=True)
    try:
        if path is None:
            os.mkfifo(result, 0o600)
        yield result
    finally:
        if path is None:
            os.unlink(result)


async def main_invoke(cmd: Sequence[str], args: Namespace) -> None:
    """Runs a program using strace"""
    access_trace_file_path = Path("~").expanduser() / f"access-log-{cmd[0]}-{datetime.now().strftime('%Y.%m.%d-%H%M')}.log"
    with open(access_trace_file_path, "w") as outfile:
        print(access_trace_file_path)
        outfile.write(f"{' '.join(cmd)}\n")

        with strace_output_path(None) as strace_output_file_path:
            full_cmd = (
                "strace",
                "--trace=openat",
                "--follow-forks",
                "--columns=0",
                "-o",
                f"{strace_output_file_path}",
                *cmd,
            )
            process = await create_subprocess_exec(*full_cmd, stdout=PIPE, stderr=PIPE)
            try:
                await gather(
                    buffer_stream(process.stdout, sys.stdout),
                    buffer_stream(process.stderr, sys.stderr),
                    process_strace_lines(strace_output_file_path, outfile),
                    process.wait(),
                )
            except (KeyboardInterrupt, CancelledError):
                pass
            finally:
                try:
                    process.terminate()
                except ProcessLookupError:
                    pass
            raise SystemExit(process.returncode)


def main() -> int:
    """Main entrypoint"""
    #args, command = parse_args()
    args, command = {}, sys.argv[1:]

    run(main_invoke(command, args))


if __name__ == "__main__":
    main()

